@@ -13,6 +13,8 @@
 #include "externalcommandhelper.h"
 #include "externalcommand_whitelist.h"
 
+#include <filesystem>
+
 #include <QtDBus>
 
 #include <QCoreApplication>
@@ -87,7 +89,7 @@ bool ExternalCommandHelper::readData(const QString& sourceDevice, QByteArray& bu
 
     if (size != buffer.size()) {
         qCritical() << xi18n("Could not read from device <filename>%1</filename>.", sourceDevice);
-         return false;
+        return false;
     }
 
     return true;
@@ -103,7 +105,7 @@ bool ExternalCommandHelper::writeData(const QString &targetDevice, const QByteAr
 {
     QFile device(targetDevice);
 
-    auto flags = QIODevice::WriteOnly | QIODevice::Unbuffered | QIODevice::Append;
+    auto flags = QIODevice::WriteOnly | QIODevice::Unbuffered;
     if (!device.open(flags)) {
         qCritical() << xi18n("Could not open device <filename>%1</filename> for writing.", targetDevice);
         return false;
@@ -153,40 +155,53 @@ bool ExternalCommandHelper::CreateFile(const QString &filePath, const QByteArray
 }
 
 // If targetDevice is empty then return QByteArray with data that was read from disk.
-QVariantMap ExternalCommandHelper::CopyBlocks(const QString& sourceDevice, const qint64 sourceFirstByte, const qint64 sourceLength, const QString& targetDevice, const qint64 targetFirstByte, const qint64 blockSize)
+QVariantMap ExternalCommandHelper::CopyBlocks(const QString& sourceDevice, const qint64 sourceOffset, const qint64 sourceLength, const QString& targetDevice, const qint64 targetOffset, const qint64 blockSize)
 {
     if (!isCallerAuthorized()) {
-        return QVariantMap();
+        return {};
     }
 
     // Avoid division by zero further down
     if (!blockSize) {
-        return QVariantMap();
+        return {};
     }
 
     // Prevent some out of memory situations
-    constexpr qint64 MiB = 1 << 30;
     if (blockSize > 100 * MiB) {
-        return QVariantMap();
-    }
-    if (targetDevice.isEmpty() && sourceLength > MiB) {
-        return QVariantMap();
+        return {};
     }
 
     QVariantMap reply;
     reply[QStringLiteral("success")] = true;
 
-    const qint64 blocksToCopy = sourceLength / blockSize;
-    qint64 readOffset = sourceFirstByte;
-    qint64 writeOffset = targetFirstByte;
-    qint32 copyDirection = 1;
+    // This enum specified whether individual blocks are moved left or right
+    // When partition is moved to the left, we start with the leftmost block,
+    // and move it further left, then second leftmost block and so on.
+    // But when we move partition to the right, we start with rightmost block.
+    // To account for this difference, we introduce CopyDirection variable which takes
+    // care of some of the differences between these two cases.
+    enum CopyDirection : qint8 {
+        Left = 1,
+        Right = -1,
+    };
+    qint8 copyDirection = targetOffset > sourceOffset ? CopyDirection::Right : CopyDirection::Left;
 
-    if (targetFirstByte > sourceFirstByte) {
-        readOffset = sourceFirstByte + sourceLength - blockSize;
-        writeOffset = targetFirstByte + sourceLength - blockSize;
-        copyDirection = -1;
+    // Let readOffset (r) and writeOffset (w) be the offsets of the first block that we move.
+    // When we move data to the left:
+    // ______target______         ______source______
+    // r                     <-   w=================
+    qint64 readOffset = sourceOffset;
+    qint64 writeOffset = targetOffset;
+
+    // When we move data to the right, we start moving data from the last block
+    // ______source______         ______target______
+    // =================r    ->                    w
+    if (copyDirection == CopyDirection::Right) {
+        readOffset = sourceOffset + sourceLength - blockSize;
+        writeOffset = targetOffset + sourceLength - blockSize;
     }
 
+    const qint64 blocksToCopy = sourceLength / blockSize;
     const qint64 lastBlock = sourceLength % blockSize;
 
     qint64 bytesWritten = 0;
@@ -199,13 +214,13 @@ QVariantMap ExternalCommandHelper::CopyBlocks(const QString& sourceDevice, const
     timer.start();
 
     QString reportText = xi18nc("@info:progress", "Copying %1 blocks (%2 bytes) from %3 to %4, direction: %5.", blocksToCopy,
-                                              sourceLength, readOffset, writeOffset, copyDirection == 1 ? i18nc("direction: left", "left")
+                                              sourceLength, readOffset, writeOffset, copyDirection == CopyDirection::Left ? i18nc("direction: left", "left")
                                               : i18nc("direction: right", "right"));
     Q_EMIT report(reportText);
 
     bool rval = true;
 
-    while (blocksCopied < blocksToCopy && !targetDevice.isEmpty()) {
+    while (blocksCopied < blocksToCopy) {
         if (!(rval = readData(sourceDevice, buffer, readOffset + blockSize * blocksCopied * copyDirection, blockSize)))
             break;
 
@@ -231,17 +246,14 @@ QVariantMap ExternalCommandHelper::CopyBlocks(const QString& sourceDevice, const
     if (rval && lastBlock > 0) {
         Q_ASSERT(lastBlock < blockSize);
 
-        const qint64 lastBlockReadOffset = copyDirection > 0 ? readOffset + blockSize * blocksCopied : sourceFirstByte;
-        const qint64 lastBlockWriteOffset = copyDirection > 0 ? writeOffset + blockSize * blocksCopied : targetFirstByte;
+        const qint64 lastBlockReadOffset = copyDirection == CopyDirection::Left ? readOffset + blockSize * blocksCopied : sourceOffset;
+        const qint64 lastBlockWriteOffset = copyDirection == CopyDirection::Left ? writeOffset + blockSize * blocksCopied : targetOffset;
         reportText = xi18nc("@info:progress", "Copying remainder of block size %1 from %2 to %3.", lastBlock, lastBlockReadOffset, lastBlockWriteOffset);
         Q_EMIT report(reportText);
         rval = readData(sourceDevice, buffer, lastBlockReadOffset, lastBlock);
 
         if (rval) {
-            if (targetDevice.isEmpty())
-                reply[QStringLiteral("targetByteArray")] = buffer;
-            else
-                rval = writeData(targetDevice, buffer, lastBlockWriteOffset);
+            rval = writeData(targetDevice, buffer, lastBlockWriteOffset);
         }
 
         if (rval) {
@@ -257,7 +269,29 @@ QVariantMap ExternalCommandHelper::CopyBlocks(const QString& sourceDevice, const
     return reply;
 }
 
-bool ExternalCommandHelper::WriteData(const QByteArray& buffer, const QString& targetDevice, const qint64 targetFirstByte)
+QByteArray ExternalCommandHelper::ReadData(const QString& device, const qint64 offset, const qint64 length)
+{
+    if (!isCallerAuthorized()) {
+        return {};
+    }
+
+    if (length > MiB) {
+        return {};
+    }
+    if (!std::filesystem::is_block_file(device.toStdString())) {
+        qWarning() << "Not a block device";
+        return {};
+    }
+
+    QByteArray buffer;
+    bool rval = readData(device, buffer, offset, length);
+    if (rval) {
+        return buffer;
+    }
+    return QByteArray();
+}
+
+bool ExternalCommandHelper::WriteData(const QByteArray& buffer, const QString& targetDevice, const qint64 targetOffset)
 {
     if (!isCallerAuthorized()) {
         return false;
@@ -266,13 +300,13 @@ bool ExternalCommandHelper::WriteData(const QByteArray& buffer, const QString& t
     if ( targetDevice.left(5) != QStringLiteral("/dev/") )
         return false;
 
-    return writeData(targetDevice, buffer, targetFirstByte);
+    return writeData(targetDevice, buffer, targetOffset);
 }
 
 QVariantMap ExternalCommandHelper::RunCommand(const QString& command, const QStringList& arguments, const QByteArray& input, const int processChannelMode)
 {
     if (!isCallerAuthorized()) {
-        return QVariantMap();
+        return {};
     }
     QTextCodec::setCodecForLocale(QTextCodec::codecForName("UTF-8"));
     QVariantMap reply;
@@ -285,9 +319,8 @@ QVariantMap ExternalCommandHelper::RunCommand(const QString& command, const QStr
 
     // Compare with command whitelist
     QString basename = command.mid(command.lastIndexOf(QLatin1Char('/')) + 1);
-    if (std::find(std::begin(allowedCommands), std::end(allowedCommands), basename) == std::end(allowedCommands)) {
+    if (allowedCommands.find(basename) == allowedCommands.end()) { // TODO: C++20: replace with contains
         qInfo() << command <<" command is not one of the whitelisted command";
-        qApp->quit();
         reply[QStringLiteral("success")] = false;
         return reply;
     }
@@ -332,6 +365,9 @@ bool ExternalCommandHelper::isCallerAuthorized()
 
     // Cache successful authentication requests, so that clients don't need
     // to authenticate multiple times during long partitioning operations.
+    // auth_admin_keep is not used intentionally because with current architecture
+    // it might lead to data loss if user cancels sfdisk partition boundary adjustment
+    // after partition data was moved.
     if (m_serviceWatcher->watchedServices().contains(message().service())) {
         return true;
     }
